Unit protocolstructs;

//  ////////////////////////////////////////////////////////////////////////////
//
//  Nord Modular Midi Protocol 3.03 Library
//  Copyright (C) 2003 Marcus Andersson
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//  ////////////////////////////////////////////////////////////////////////////
//
//  Translated into Pascal by Jan Punter, meant to be compiled under Delphi 7.
//
//  Modifications CopyRight (c) 2003 Jan Punter
//
//  Distributed under the same conditions as the original code by Narcus
//  Andersson.
//
//  Changes from original :
//
//    Removed MidiDriver, replaced this by AcceptDriverData and
//    OnDriverData members for TNMProtocol. AcceptDriverData should be called by
//    the driver when it has received data. OnDriverData (if assigned) willl be
//    called by TNMProtocol when it wants to send data to a driver.
//
//    2003-11-22
//
//  Relative positions in EndPositions changed into absolute positions,
//  could not get it right with large sections otherwise.
//
//  Changed the Send mechanism to account better for retries and
//  unexpected replies.
//
//  ////////////////////////////////////////////////////////////////////////////


Interface

Uses

  Classes, SysUtils, Contnrs, Windows, mmSystem, Globals, pdlstructs,
  patchstructs;

Type

  EProtocol       = Class( PdlException);
  EMidiMessage    = Class( EProtocol);
  ENotImplemented = Class( EProtocol);


  TByteList           = Class;
  TMidiMessage        = Class;
  TEnqueuedPacket     = Class;

  TOnDriverData  = Procedure( aSender: TObject; aData: TByteList   ) Of Object;
  TOnMidiMessage = Procedure( aSender: TObject; aData: TMidiMessage) Of Object;
  TOnUnparsable  = Procedure( aSender: TObject; aData: TByteList   ) Of Object;


  TByteList = Class
  Private
    FData : Array Of Byte;
  Private
    Function    GetSize: Integer;
    Procedure   SetSize( aValue: Integer);
    Function    GetItem( anIndex: Integer): Byte;
    Procedure   SetItem( anIndex: Integer; aValue: Byte);
    Function    GetAsString: String;
  Public
    Constructor Create;
    Destructor  Destroy;                                               Override;
    Procedure   Clear;
    Procedure   Push( aValue: Byte);
    Function    Pop: Byte;
    Function    Head: Byte;
  Public
    Property    Size: Integer                  Read GetSize     Write SetSize;
    Property    Items[ anIndex: Integer]: Byte Read GetItem     Write SetItem; Default;
    Property    AsString: String               Read GetAsString;
  End;

  TIntList = Class
  Private
    FData : Array Of Integer;
  Private
    Function    GetSize: Integer;
    Procedure   SetSize( aValue: Integer);
    Function    GetItem( anIndex: Integer): Integer;
  Public
    Constructor Create;
    Destructor  Destroy;                                               Override;
    Procedure   Clear;
    Procedure   Push( aValue: Integer);
    Function    Pop: Integer;
    Function    Head: Integer;
  Public
    Property    Size: Integer Read GetSize Write SetSize;
    Property    Items[ anIndex: Integer]: Integer Read GetItem;
  End;

  TBitStreamList = Class( TPdlObjectList)
  Private
    Function    GetItem( anIndex: Integer): TBitStream;
  Public
    Procedure   Push( aValue: TBitStream);
    Function    Head: TBitStream;
  Public
    Property    Items[ anIndex: Integer]: TBitStream Read GetItem;
  End;

  TPacketQueue = Class( TObjectQueue)
  Public
    Function    Push( aValue: TEnqueuedPacket): TEnqueuedPacket;
    Function    Pop : TEnqueuedPacket;
    Function    Peek: TEnqueuedPacket;
    Procedure   Drop;
    Procedure   Clear;
  End;

  TMidiMessage = Class
  Protected
    Fcc           : Integer;
    FSlot         : Integer;
    FIsReply      : Boolean;
    FExpectsReply : Boolean;
  Private
    Class Function    GetPdlFile : String;
    Class Procedure   SetPdlFile( Const aValue: String);
    Class Function    GetProtocol : TProtocol;
    Class Procedure   SetProtocol( aValue: TProtocol);
    Class Function    GetPacketParser: TPacketParser;
    Class Procedure   SetPacketParser( aValue: TPacketParser);
  Protected
    Procedure   GetBitStream( anIntStream: TIntStream; aBitStream: TBitStream); Overload;
    Procedure   AppendChecksum( anIntStream: TIntStream);
    Function    ExtractName( aName: TPacket): String;
    Constructor Create;
  Public
    Class Function    CreateMsg( aBitStream: TBitStream): TMidiMessage;
    Class Function    CalculateCheckSum( aBitStream: TBitStream): Integer;
    Class Function    ChecksumIsCorrect( aBitStream: TBitStream): Boolean;
  Public
    Destructor  Destroy;                                               Override;
    Procedure   UsePdlFile( Const aFileName: String; aTracer: TTracer);
    Procedure   GetBitStream( aBitStreamList: TBitStreamList); Overload; Virtual; Abstract;
    Function    ExpectsReply: Boolean;                                  Virtual;
    Function    IsReply: Boolean;                                       Virtual;
  Private
    Property    PdlFile     : String        Read GetPdlFile      Write SetPdlFile;
    Property    Protocol    : TProtocol     Read GetProtocol     Write SetProtocol;
    Property    PacketParser: TPacketParser Read GetPacketParser Write SetPacketParser;
  Public
    Property    Slot        : Integer       Read FSlot           Write FSlot;
  End;


  TSender = (
    sePC,       // 0
    seModular   // 1
  );

  TIamMessage = Class( TMidiMessage)
  Private
    FSender      : TSender;
    FVersionHigh : Integer;
    FVersionLow  : Integer;
  Public
    Constructor Create;                                                Overload;
    Constructor Create( aPacket: TPacket);                             Overload;
    Procedure   SetVersion( aVersionHigh, aVersionLow: Integer);
    Procedure   GetBitStream( aBitStreamList: TBitStreamList);         Override;
  Public
    Property    Sender      : TSender Read FSender;
    Property    VersionHigh : Integer Read FVersionHigh Write FVersionHigh;
    Property    VersionLow  : Integer Read FVersionLow  Write FVersionLow;
  End;

  TVoiceCountMessage = Class( TMidiMessage)
  Private
    FVoices : Array[ 0 .. 3] Of Integer;
  Private
    Function    GetVoiceCount( anIndex: Integer): Integer;
  Public
    Constructor Create;                                                Overload;
    Constructor Create( aPacket: TPacket);                             Overload;
    Procedure   GetBitStream( aBitStreamList: TBitStreamList);         Override;
  Public
    Property    VoiceCount[ anIndex: Integer]: Integer Read GetVoiceCount;
  End;

  TSlotsSelectedMessage = Class( TMidiMessage)
  Private
    // Fpid      : Integer;
    Fsc       : Integer;
    FSelected : Array[ 0 .. 3] Of Boolean;
  Private
    Function    GetIsSelected( anIndex: Integer): Boolean;
    Procedure   SetIsSelected( anIndex: Integer; aValue: Boolean);
  Public
    Constructor Create;                                                Overload;
    Constructor Create( aPacket: TPacket);                             Overload;
    Procedure   GetBitStream( aBitStreamList: TBitStreamList);         Override;
  Public
    Property    IsSelected[ anIndex: Integer]: Boolean Read GetIsSelected Write SetIsSelected;
  End;

  TSlotActivatedMessage = Class( TMidiMessage)
  Private
    // Fpid        : Integer;
    Fsc         : Integer;
    FActiveSlot : Integer;
  Public
    Constructor Create;                                                Overload;
    Constructor Create( aPacket: TPacket);                             Overload;
    Procedure   GetBitStream( aBitStreamList: TBitStreamList);         Override;
  Public
    Property    ActiveSlot: Integer Read FActiveSlot Write FActiveSlot;
  End;

  TNewPatchInSlotMessage = Class( TMidiMessage)
  Private
    FPid : Integer;
  Public
    Constructor Create;                                                Overload;
    Constructor Create( aPacket: TPacket);                             Overload;
    Procedure   GetBitStream( aBitStreamList: TBitStreamList);         Override;
  Public
    Property    Pid: Integer Read FPid Write FPid;
  End;

  TLightMessage = Class( TMidiMessage)
  Private
    Fpid        : Integer;
    FStartIndex : Integer;
    FLights     : Array[ 0 .. 19] Of Integer;
  Private
    Function    GetLights( anIndex: Integer): Integer;
  Public
    Constructor Create;                                                Overload;
    Constructor Create( aPacket: TPacket);                             Overload;
    Procedure   GetBitStream( aBitStreamList: TBitStreamList);         Override;
  Public
    Property    Pid        : Integer Read Fpid;
    Property    StartIndex : Integer Read FStartIndex;
    Property    Lights[ anIndex: Integer]: Integer Read GetLights;
  End;

  TPatchListMessage = Class( TMidiMessage)
  Private
    FSection   : Integer;
    FPosition  : Integer;
    FEndOfList : Integer;
    FNames     : TStringList;
  Private
    Function    GetEndOfList: Boolean;
  Public
    Constructor Create;                                                Overload;
    Constructor Create( aSection, aPosition: Integer);                 Overload;
    Constructor Create( aPacket: TPacket);                             Overload;
    Destructor  Destroy;                                               Override;
    Procedure   GetBitStream( aBitStreamList: TBitStreamList);         Override;
    Function    EndOfSection ( Const aName: String): Boolean;
    Function    EmptyPosition( Const aName: String): Boolean;
  Public
    Property    Section  : Integer     Read FSection;
    Property    Position : Integer     Read FPosition;
    Property    Names    : TStringList Read FNames;
    Property    EndOfList: Boolean     Read GetEndOfList;
  End;

  TGetPatchListMessage = Class( TMidiMessage)
  Private
    Fpp       : Integer;
    Fssc      : Integer;
    FSection  : Integer;
    FPosition : Integer;
  Public
    Constructor Create( aPacket: TPacket);                             Overload;
    Constructor Create( aSection, aPosition: Integer);                 Overload;
    Procedure   GetBitStream( aBitStreamList: TBitStreamList);         Override;
  End;

  TAckMessage = Class( TMidiMessage)
  Private
    Fpid1    : Integer;
    Fpid2    : Integer;
    FAckType : Integer; // Type is a reserved word in Pascal
  Public
    Constructor Create;                                                Overload;
    Constructor Create( aPacket: TPacket);                             Overload;
    Procedure   GetBitStream( aBitStreamList: TBitStreamList);         Override;
  Public
    Property    pid1 : Integer Read Fpid1 Write Fpid1;
    Property    pid2 : Integer Read Fpid2 Write Fpid2;
  End;

  TGetPatchMessage = Class( TMidiMessage)
  Private
    FPid : Integer;
  Public
    Constructor Create( aSlot, aPid: Integer);                         Overload;
    Constructor Create( aPacket: TPacket);                             Overload;
    Procedure   GetBitStream( aBitStreamList: TBitStreamList);         Override;
  End;

  TPatchMessage = Class( TMidiMessage)
  Private
    FPatch       : TPatch;
    FPatchStream : TBitStream;
    FPid         : Integer;
  Private
    Class Function PatchPdlFile : String;
    Class Function PatchParser  : TPacketParser;
  Private
    Procedure   Init;
    Procedure   AppendName( Const aName: String; aPatchStream: TIntStream);
    Procedure   StoreEndPosition( anIntStream: TIntStream; EndPositions: TIntList);
    Function    GetModule(
      aPatch        : TPatch;
      aSection      : TTModuleSection;
      anIndex       : Integer;
      Const aContext: String
    ): TModule;
  Public
    Constructor Create;                                               Overload;
    Constructor Create( aPatch : TPatch );                             Overload;
    Destructor  Destroy;                                               Override;
    Procedure   GetBitStream( aBitStreamList: TBitStreamList);         Override;
    Procedure   Append( aPacket: TPacket);
    Procedure   GetPatch( aPatch: TPatch);
  Public
    Class Procedure UsePdlFile( Const aFileName: String; aTracer: TTracer = Nil);
  Public
    Property    Patch: TPatch  Read FPatch;
    Property    Pid  : Integer Read FPid;
  End;

  TEnqueuedPacket = Class
  Private
    FContent      : TByteList;
    FExpectsReply : Boolean;
  Public
    Constructor Create( aContent: TByteList; anExpectsReply: Boolean);
    Destructor  Destroy;                                               Override;
  Public
    Property    Content     : TByteList Read FContent;
    Property    ExpectsReply: Boolean   Read FExpectsReply;
  End;

  TNMProtocolDriver = Class
  Private
    FOnDriverData    : TOnDriverData;               // Transmit to driver
    FOnMidiMessage   : TOnMidiMessage;              // Application midi message handler
    FOnUnparsable    : TOnUnparsable;
    FReceiveState    : ( rsIdle, rsInSysex);
    FRxBuffer        : TByteList;
    FTxBuffer        : TPacketQueue;
    FTimeOutTime     : DWORD;                       // ms, 0 for no timeout
    FTimeoutInterval : DWORD;
  Private
    Procedure   NotifyListeners( aMessage: TMidiMessage);
    Procedure   NotifyUnparsable( aData: TByteList);
    Function    ReplyExpected: Boolean;
    Function    TimeoutOccured: Boolean;
    Procedure   StartTimeout( aTimeInms: DWORD);
    Procedure   StopTimeout;
  Protected
    Procedure   AcceptMidiMessage( aMidiMsg: TMidiMessage);
    Procedure   AcceptMidiPacket( aData: TByteList);
    Procedure   DoTransmit;
  Public
    Constructor Create;
    Destructor  Destroy;                                               Override;
    Procedure   ClearTransmitter;
    Procedure   HeartBeat;
    Procedure   AcceptDriverData( aData: TByteList);
    Procedure   AcceptRawDriverData( aData: PBytes; aSize: Integer);
    Procedure   Send( aMessage: TMidiMessage);
  Protected
    Property    TimeoutTime: DWORD Read FTimeoutTime Write FTimeoutTime;
  Public
    Property    OnDriverData   : TOnDriverData  Read FOnDriverData    Write FOnDriverData;
    Property    OnMidiMessage  : TOnMidiMessage Read FOnMidiMessage   Write FOnMidiMessage;
    Property    OnUnparsable   : TOnUnparsable  Read FOnUnparsable    Write FOnUnparsable;
    Property    TimeoutInterval: DWORD          Read FTimeoutInterval Write FTimeoutInterval;
  End;


  Function SenderToStr( aSender: TSender): String;


Implementation


  Function SenderToStr( aSender: TSender): String;
  Begin
    Case aSender Of
      sePC      : Result := 'PC';
      seModular : Result := 'Modular';
    End;
  End;

Const

  SYSEX_START = $f0;
  SYSEX_END   = $f7;


{ ========
  TByteList = Class
  Private
    FData : Array Of Byte;
  Public
    Property    Size: Integer                  Read GetSize     Write SetSize;
    Property    Items[ anIndex: Integer]: Byte Read GetItem     Write SetItem; Default;
    Property    AsString: String               Read GetAsString;
  Private
}

    Function    TByteList.GetSize: Integer;
    Begin
      Result := Length( FData);
    End;

    Procedure   TByteList.SetSize( aValue: Integer);
    Begin
      SetLength( FData, aValue);
    End;

    Function    TByteList.GetItem( anIndex: Integer): Byte;
    Begin
      Result := FData[ anIndex];
    End;

    Procedure   TByteList.SetItem( anIndex: Integer; aValue: Byte);
    Begin
      FData[ anIndex] := aValue;
    End;

    Function    TByteList.GetAsString: String;
    Var
      i : Integer;
    Begin
      Result := '[ ';
      For i := 0 To Size - 1 Do
        Result := Result + Format( ' %.2x', [ Items[ i]]);
      Result := Result + ']';
    End;

//  Public

    Constructor TByteList.Create;
    Begin
      Inherited;
      Clear;
    End;

    Destructor  TByteList.Destroy; // Override;
    Begin
      Clear;
      Inherited;
    End;

    Procedure   TByteList.Clear;
    Begin
      Size := 0;
    End;

    Procedure   TByteList.Push( aValue: Byte);
    Begin
      Size := Size + 1;
      FData[ Size - 1] := aValue;
    End;

    Function    TByteList.Pop: Byte;
    Begin
      Result := FData[ Size - 1];
      Size := Size - 1;
    End;

    Function    TByteList.Head: Byte;
    Begin
      Result := FData[ 0];
    End;


{ ========
  TIntList = Class
  Private
    FData : Array Of Integer;
  Public
    Property    Size: Integer Read GetSize Write SetSize;
    Property    Items[ anIndex: Integer]: Integer Read GetItem;
  Private
}

    Function    TIntList.GetSize: Integer;
    Begin
      Result := Length( FData);
    End;

    Procedure   TIntList.SetSize( aValue: Integer);
    Begin
      SetLength( FData, aValue);
    End;

    Function    TIntList.GetItem( anIndex: Integer): Integer;
    Begin
      Result := FData[ anIndex];
    End;

//  Public

    Constructor TIntList.Create;
    Begin
      Inherited;
      Clear;
    End;

    Destructor  TIntList.Destroy; // Override;
    Begin
      Clear;
      Inherited;
    End;

    Procedure   TIntList.Clear;
    Begin
      Size := 0;
    End;

    Procedure   TIntList.Push( aValue: Integer);
    Begin
      Size := Size + 1;
      FData[ Size - 1] := aValue;
    End;

    Function    TIntList.Pop: Integer;
    Begin
      Result := FData[ Size - 1];
      Size := Size - 1;
    End;

    Function    TIntList.Head: Integer;
    Begin
      Result := FData[ 0];
    End;


{ ========
  TBitStreamList = Class( TPdlObjectList)
  Public
    Property    Items[ anIndex: Integer]: TBitStream Read GetItem;
  Private
}

    Function    TBitStreamList.GetItem( anIndex: Integer): TBitStream;
    Begin
      Result := Inherited Items[ anIndex] As TBitStream;
    End;

//  Public

    Procedure   TBitStreamList.Push( aValue: TBitStream);
    Begin
      Inherited Push( aValue);
    End;

    Function    TBitStreamList.Head: TBitStream;
    Begin
      Result := Inherited Head As TBitStream;
    End;


{ ========
  TPacketQueue = Class( TObjectQueue)
  Public
}

    Function    TPacketQueue.Push( aValue: TEnqueuedPacket): TEnqueuedPacket;
    Begin
      Result := Inherited Push( aValue) As TEnqueuedPacket;
    End;

    Function    TPacketQueue.Pop : TEnqueuedPacket;
    Begin
      Result := Inherited Pop As TEnqueuedPacket;
    End;

    Function    TPacketQueue.Peek: TEnqueuedPacket;
    Begin
      If Count = 0
      Then Result := Nil
      Else Result := Inherited Peek As TEnqueuedPacket;
    End;

    Procedure   TPacketQueue.Drop;
    Begin
      If Count > 0
      Then Pop.Free;
    End;

    Procedure   TPacketQueue.Clear;
    Begin
      While Count > 0 Do
        Pop.Free;
    End;

{ ========
  TMidiMessage = Class
  Protected
    Fcc           : Integer;
    FSlot         : Integer;
    FIsReply      : Boolean;
    FExpectsReply : Boolean;
  Private
    Property    PdlFile     : String        Read GetPdlFile      Write SetPdlFile;
    Property    Protocol    : TProtocol     Read GetProtocol     Write SetProtocol;
    Property    PacketParser: TPacketParser Read GetPacketParser Write SetPacketParser;
  Public
    Property    Slot: Integer Read FSlot Write FSlot;
  Public
    Procedure   GetBitStream( aBitStreamList: TBitStreamList); Overload; Virtual; Abstract;
  Private
}

Var

// Poor man's class (static) variables for TMidiMessage :

  GTMidiMessageProtocol     : TProtocol     = Nil;  // Statically assigned.
  GTMidiMessagePacketParser : TPacketParser = Nil;  // Statically assigned.
  GTMidiMessagePdlFile      : String;               // Dynamically assigned in
                                                    //   Initialization section.

    Class Function    TMidiMessage.GetPdlFile : String;
    Begin
      Result := GTMidiMessagePdlFile;
    End;

    Class Procedure   TMidiMessage.SetPdlFile( Const aValue: String);
    Begin
      GTMidiMessagePdlFile := aValue;
      SetProtocol( TProtocol.Create);
      With GetProtocol Do
      Begin
        LoadPdlFile( GTMidiMessagePdlFile);
        SetPacketParser( PacketParser[ 'Sysex']);
      End;
    End;

    Class Function    TMidiMessage.GetProtocol : TProtocol;
    Begin
      Result := GTMidiMessageProtocol;
    End;

    Class Procedure   TMidiMessage.SetProtocol( aValue: TProtocol);
    Begin
      If Assigned( GTMidiMessageProtocol)
      Then FreeAndNil( GTMidiMessageProtocol);
      GTMidiMessageProtocol := aValue;
    End;

    Class Function    TMidiMessage.GetPacketParser: TPacketParser;
    Begin
      Result := GTMidiMessagePacketParser;
    End;

    Class Procedure   TMidiMessage.SetPacketParser( aValue: TPacketParser);
    Begin
      GTMidiMessagePacketParser := aValue;
    End;

//  Protected

    Procedure   TMidiMessage.GetBitStream( anIntStream: TIntStream; aBitStream: TBitStream); // Overload;
    Var
      Success : Boolean;
    Begin
      Success := PacketParser.Generate( anIntStream, aBitStream);
      If Not Success Or anIntStream.HasAvailable( 1)
      Then Raise EMidiMessage.Create( 'Information mismatch in Generate', Ord( Success));
    End;

    Procedure   TMidiMessage.AppendChecksum( anIntStream: TIntStream);
    Var
      aSum       : Integer;
      aBitStream : TBitStream;
    Begin
      aSum := 0;
      anIntStream.Append( aSum);  // pff ... ok ...
      aBitStream := TBitStream.Create;
      Try
        GetBitStream( anIntStream, aBitStream);
        aSum := CalculateCheckSum( aBitStream);
        With anIntStream Do
        Begin
          Size := Size - 1;
          Append( aSum);
          Position := 0;
        End;
      Finally
        aBitStream.Free;
      End;
    End;

    Function    TMidiMessage.ExtractName( aName: TPacket): String;
    Var
      aChars : TVariableList;
      i      : Integer;
    Begin
      Result := '';
      aChars := aName.VariableList[ 'chars'];
      If Assigned( aChars)
      Then
        With aChars Do
          For i := 0 To Size - 1 Do
            If Items[ i] <> 0
            Then Result := Result + Char( Items[ i]);
    End;

    Constructor TMidiMessage.Create; // Overload
    Begin
      Inherited;
      FExpectsReply := False;
      FIsReply      := False;
      If Not Assigned( Protocol)
      Then Begin
        Protocol := TProtocol.Create;
        Protocol.LoadPdlFile( PdlFile);
        PacketParser := Protocol.PacketParser[ 'Sysex'];
      End;
    End;

//  Public

    Class Function TMidiMessage.CreateMsg( aBitStream: TBitStream): TMidiMessage; // Overload;
    Var
      aPacket : TPacket;
      Success : Boolean;
    Begin
      Result  := Nil;
      aPacket := TPacket.Create;
      Success := GTMidiMessagePacketParser.Parse( aBitStream, aPacket);
      aBitStream.Position := 0;
      If Success
      Then Begin
        If aPacket.Contains( 'IAm')
        Then Result := TIamMessage.Create( aPacket)
        Else If ChecksumIsCorrect( aBitStream)
        Then Begin
          If      aPacket.Contains( 'VoiceCount')        Then Result := TVoiceCountMessage    .Create( aPacket)
          Else If aPacket.Contains( 'SlotsSelected')     Then Result := TSlotsSelectedMessage .Create( aPacket)
          Else If aPacket.Contains( 'SlotActivated')     Then Result := TSlotActivatedMessage .Create( aPacket)
          Else If aPacket.Contains( 'NewPartchInSlot')   Then Result := TNewPatchInSlotMessage.Create( aPacket)
          Else If aPacket.Contains( 'Lights')            Then Result := TLightMessage         .Create( aPacket)
          Else If aPacket.Contains( 'PatchListResponse') Then Result := TPatchListMessage     .Create( aPacket)
          Else If aPacket.Contains( 'ACK')               Then Result := TAckMessage           .Create( aPacket)
          Else If aPacket.Contains( 'PatchPacket')
          Then Begin
            Result := TPatchMessage.Create;
            TPatchMessage( Result).Append( aPacket);
          End
          Else {printf("unsupported packet : ")};
        End
        Else {printf("checksumerror : ")};
      End
      Else {printf( "Parse failed : ")};
      If Not Success
      Then
        While aBitStream.HasAvailable( 8) Do
          aBitStream.GetInt( 8);
          {printf("%X ", bitStream->getInt( 8))};
      {printf( "\n")}
    End;

    Class Function    TMidiMessage.CalculateCheckSum( aBitStream: TBitStream): Integer;
    Begin
      Result := 0;
      aBitStream.Position := 0;
      While( aBitStream.HasAvailable( 24)) Do
        Result := ( Result + aBitStream.GetInt( 8)) And $7f;
    End;

    Class Function    TMidiMessage.ChecksumIsCorrect( aBitStream: TBitStream): Boolean;
    Var
      aChecksum : Integer;
      bCheckSum : Integer;
    Begin
      aChecksum := CalculateCheckSum( aBitStream);
      With aBitStream Do
      Begin
        Position  := Size - 16;
        bChecksum := GetInt( 8);
      End;
      Result := aChecksum = bChecksum;
      If Not Result
      Then {printf("Checksum error %d <> %d", aChecksum, bChecksum)};
    End;

//  Public

    Destructor  TMidiMessage.Destroy; // Override;
    Begin
      Inherited;
    End;

    Procedure   TMidiMessage.UsePdlFile( Const aFileName: String; aTracer: TTracer);
    Begin
      PdlFile := aFileName;
      Protocol.UseTracer( aTracer);
    End;

    Function    TMidiMessage.ExpectsReply: Boolean; // Virtual;
    Begin
      Result := FExpectsReply;
    End;

    Function    TMidiMessage.IsReply: Boolean; // Virtual;
    Begin
      Result := FIsReply;
    End;

{ ========
  TIamMessage = Class( TMidiMessage)
  TIamMessage = Class( TMidiMessage)
  Private
    FSender      : TSender;
    FVersionHigh : Integer;
    FVersionLow  : Integer;
  Public
    Property    Sender      : TSender Read FSender;
    Property    VersionHigh : Integer Read FVersionHigh Write FVersionHigh;
    Property    VersionLow  : Integer Read FVersionLow  Write FVersionLow;
  Public
}

    Constructor TIamMessage.Create; // Overload;
    Begin
      Inherited Create;
      Slot          := 0;
      FSender       := sePC;
      VersionHigh   := 0;
      VersionLow    := 0;
      FExpectsReply := True;
      FIsReply      := True;
    End;

    Constructor TIamMessage.Create( aPacket: TPacket); // Overload;
    Begin
      Create;
      Fcc           :=          aPacket.Variable[ 'cc'              ];
      Slot          :=          aPacket.Variable[ 'slot'            ];
      FSender       := TSender( aPacket.Variable[ 'data:sender'     ]);
      VersionHigh   :=          aPacket.Variable[ 'data:versionHigh'];
      VersionLow    :=          aPacket.Variable[ 'data:versionLow' ];
      FExpectsReply := True;
      FIsReply      := True;
    End;

    Procedure   TIamMessage.SetVersion( aVersionHigh, aVersionLow: Integer);
    Begin
      VersionHigh := aVersionHigh;
      VersionLow  := aVersionLow;
    End;

    Procedure   TIamMessage.GetBitStream( aBitStreamList: TBitStreamList); // Override;
    Var
      anIntStream: TIntStream;
      aBitStream : TBitStream;
    Begin
      anIntStream := TIntStream.Create;
      With anIntStream Do
      Try
        Append( Fcc);
        Append( Slot);
        Append( Integer( Sender));
        Append( VersionHigh);
        Append( VersionLow);
        aBitStream := TBitStream.Create;
        GetBitStream( anIntStream, aBitStream);
        aBitStreamList.Push( aBitStream);
      Finally
        Free;
      End;
    End;


{ ========
  TVoiceCountMessage = Class( TMidiMessage)
  Private
    FVoices : Array[ 0 .. 3] Of Integer;
  Public
    Property    VoiceCount[ anIndex: Integer]: Integer Read GetVoiceCount;
  Private
}

    Function    TVoiceCountMessage.GetVoiceCount( anIndex: Integer): Integer;
    Begin
      Result := FVoices[ anIndex];
    End;

//  Public

    Constructor TVoiceCountMessage.Create; // Overload;
    Var
      i : Integer;
    Begin
      Inherited Create;
      FCc  := 0;
      Slot := 0;
      For i := 0 To 3 Do
        FVoices[ i] := 0;
    End;

    Constructor TVoiceCountMessage.Create( aPacket: TPacket); // Overload;
    Begin
      Create;
      Fcc         := aPacket.Variable[ 'cc'          ];
      Slot        := aPacket.Variable[ 'slot'        ];
      FVoices[ 0] := aPacket.Variable[ 'data:data:c0'];
      FVoices[ 1] := aPacket.Variable[ 'data:data:c1'];
      FVoices[ 2] := aPacket.Variable[ 'data:data:c2'];
      FVoices[ 3] := aPacket.Variable[ 'data:data:c3'];
    End;

    Procedure   TVoiceCountMessage.GetBitStream( aBitStreamList: TBitStreamList); // Override;
    Begin
      Raise EMidiMessage.Create( 'TVoiceCount.GetBitStream not implemented', 0);
    End;


{ ========
  TSlotsSelectedMessage = Class( TMidiMessage)
  Private
    Fpid      : Integer;
    Fsc       : Integer;
    FSelected : Array[ 0 .. 3] Of Boolean;
  Public
    Property    IsSelected[ anIndex: Integer]: Boolean Read GetIsSelected Write SetIsSelected;
  Private
}

    Function    TSlotsSelectedMessage.GetIsSelected( anIndex: Integer): Boolean;
    Begin
      Result := FSelected[ anIndex];
    End;

    Procedure   TSlotsSelectedMessage.SetIsSelected( anIndex: Integer; aValue: Boolean);
    Begin
      FSelected[ anIndex] := aValue;
    End;

//  Public

    Constructor TSlotsSelectedMessage.Create; // Overload;
    Var
      i : Integer;
    Begin
      Inherited Create;
      FSlot  := 0;
      Fsc    := $07;
      For i := 0 To 3 Do
        FSelected[ i] := False;
      FExpectsReply := True;
    End;

    Constructor TSlotsSelectedMessage.Create( aPacket: TPacket); // Overload;
    Begin
      FSlot         :=          aPacket.Variable[ 'slot'           ];
      FSelected[ 0] := Boolean( aPacket.Variable[ 'data:data:slot0']);
      FSelected[ 1] := Boolean( aPacket.Variable[ 'data:data:slot1']);
      FSelected[ 2] := Boolean( aPacket.Variable[ 'data:data:slot2']);
      FSelected[ 3] := Boolean( aPacket.Variable[ 'data:data:slot3']);
    End;

    Procedure   TSlotsSelectedMessage.GetBitStream( aBitStreamList: TBitStreamList); // Override;
    Var
      anIntStream : TIntStream;
      aBitStream  : TBitStream;
    Begin
      anIntStream := TIntStream.Create;
      With anIntStream Do
      Try
        Append( $17  );
        Append( FSlot);
        Append( $41  );
        Append( Fsc  );
        Append( Integer( FSelected[ 0]));
        Append( Integer( FSelected[ 1]));
        Append( Integer( FSelected[ 2]));
        Append( Integer( FSelected[ 3]));
        AppendChecksum( anIntStream);
        aBitStream := TBitStream.Create;
        GetBitStream( anIntStream, aBitStream);
        aBitStreamList.Push( aBitStream);
      Finally
        Free;
      End;
    End;


{ ========
  TSlotActivatedMessage = Class( TMidiMessage)
  Private
    Fpid        : Integer;
    Fsc         : Integer;
    FActiveSlot : Integer;
  Public
    Property    ActiveSlot: Integer Read FActiveSlot Write FActiveSlot;
  Public
}

    Constructor TSlotActivatedMessage.Create; // Overload;
    Begin
      Inherited Create;
      FSlot         := 0;
      Fsc           := $09;
      ActiveSlot    := 0;
      FExpectsReply := True;
    End;

    Constructor TSlotActivatedMessage.Create( aPacket: TPacket); // Overload;
    Begin
      Create;
      FSlot      := aPacket.Variable[ 'slot'          ];
      ActiveSlot := aPacket.Variable[ 'data:data:slot'];
    End;

    Procedure   TSlotActivatedMessage.GetBitStream( aBitStreamList: TBitStreamList); // Override;
    Var
      anIntStream : TIntStream;
      aBitStream  : TBitStream;
    Begin
      anIntStream := TIntStream.Create;
      With anIntStream Do
      Try
        Append( $17);
        Append( Slot);
        Append( $41);
        Append( Fsc);
        Append( ActiveSlot);
        AppendChecksum( anIntStream);
        aBitStream := TBitStream.Create;
        GetBitStream( anIntStream, aBitStream);
        aBitStreamList.Push( aBitStream);
      Finally
        Free
      End;
    End;


{ ========
  TNewPatchInSlotMessage = Class( TMidiMessage)
  Private
    FPid : Integer;
  Public
    Property    Pid: Integer Read FPid Write FPid;
  Public
}

    Constructor TNewPatchInSlotMessage.Create; // Overload;
    Begin
      Inherited Create;
      Fcc  := $14;
      Slot := 0;
      Pid  := 0;
    End;

    Constructor TNewPatchInSlotMessage.Create( aPacket: TPacket); // Overload;
    Begin
      Create;
      Fcc  := aPacket.Variable[' cc'      ];
      Slot := aPacket.Variable[' slot'    ];
      Pid  := aPacket.Variable[' data:pid'];
    End;

    Procedure   TNewPatchInSlotMessage.GetBitStream( aBitStreamList: TBitStreamList); // Override;
    Begin
      Raise EMidiMessage.Create( 'TNewPatchInSlotMessage.getBitStream not implemented.', 0);
    End;


{ ========
  TLightMessage = Class( TMidiMessage)
  Private
    Fpid        : Integer;
    FStartIndex : Integer;
    FLights     : Array[ 0 .. 19] Of Integer;
  Public
    Property    Pid        : Integer Read Fpid;
    Property    StartIndex : Integer Read FStartIndex;
    Property    Lights[ anIndex: Integer]: Integer Read GetLights;
  Private
}

    Function    TLightMessage.GetLights( anIndex: Integer): Integer;
    Begin
      Result := FLights[ anIndex];
    End;

//  Public

    Constructor TLightMessage.Create; // Overload;
    Begin
      Inherited Create;
      Fcc         := $14;
      Slot        := 0;
      Fpid        := 0;
      FStartIndex := 0;
    End;

    Constructor TLightMessage.Create( aPacket: TPacket); // Overload;
    Begin
      Create;
      Fcc          := aPacket.Variable[ 'cc'                  ];
      Slot         := aPacket.Variable[ 'slot'                ];
      Fpid         := aPacket.Variable[ 'data:pid'            ];
      FStartIndex  := aPacket.Variable[ 'data:data:startIndex'];
      FLights[  0] := aPacket.Variable[ 'data:data:l0'        ];
      FLights[  1] := aPacket.Variable[ 'data:data:l1'        ];
      FLights[  2] := aPacket.Variable[ 'data:data:l2'        ];
      FLights[  3] := aPacket.Variable[ 'data:data:l3'        ];
      FLights[  4] := aPacket.Variable[ 'data:data:l4'        ];
      FLights[  5] := aPacket.Variable[ 'data:data:l5'        ];
      FLights[  6] := aPacket.Variable[ 'data:data:l6'        ];
      FLights[  7] := aPacket.Variable[ 'data:data:l7'        ];
      FLights[  8] := aPacket.Variable[ 'data:data:l8'        ];
      FLights[  9] := aPacket.Variable[ 'data:data:l9'        ];
      FLights[ 10] := aPacket.Variable[ 'data:data:l10'       ];
      FLights[ 11] := aPacket.Variable[ 'data:data:l11'       ];
      FLights[ 12] := aPacket.Variable[ 'data:data:l12'       ];
      FLights[ 13] := aPacket.Variable[ 'data:data:l13'       ];
      FLights[ 14] := aPacket.Variable[ 'data:data:l14'       ];
      FLights[ 15] := aPacket.Variable[ 'data:data:l15'       ];
      FLights[ 16] := aPacket.Variable[ 'data:data:l16'       ];
      FLights[ 17] := aPacket.Variable[ 'data:data:l17'       ];
      FLights[ 18] := aPacket.Variable[ 'data:data:l18'       ];
      FLights[ 19] := aPacket.Variable[ 'data:data:l19'       ];
    End;

    Procedure   TLightMessage.GetBitStream( aBitStreamList: TBitStreamList); // Override;
    Begin
      Raise EMidiMessage.Create( 'TLightMessage. getBitStream not implemented.', 0);
    End;


Const

  END_OF_SECTION = 3;
  EMPTY_POSITION = 2;

{ ========
  TPatchListMessage = Class( TMidiMessage)
  Private
    FSection   : Integer;
    FPosition  : Integer;
    FEndOfList : Integer;
    FNames     : TStringList;
  Public
    Property    Section  : Integer     Read FSection;
    Property    Position : Integer     Read FPosition;
    Property    Names    : TStringList Read FNames;
    Property    EndOfList: Boolean     Read FEndOfList;
  Private
    Function    GetEndOfList: Boolean;
}

    Function    TPatchListMessage.GetEndOfList: Boolean;
    Begin
      Result := FEndOfList <> 0;
    End;

//  Public

    Constructor TPatchListMessage.Create; // Overload;
    Begin
      Inherited Create;
      FNames := TStringList.Create;
    End;

    Constructor TPatchListMessage.Create( aSection, aPosition: Integer); // Overload;
    Begin
      Create;
      Fcc        := $17;
      Slot       := 0;
      FSection   := aSection;
      FPosition  := aPosition;
      FEndOfList := 0;
      FIsReply   := True;
    End;

    Constructor TPatchListMessage.Create( aPacket: TPacket); // Overload;
    Var
      aPatchList : TPacket;
      aList      : TPacket;
      aName      : String;
    Begin
      Slot       := aPacket.Variable[ 'slot'               ];
      aPatchList := aPacket.Packet  [ 'data:patchlist:data'];
      If Assigned( aPatchList)
      Then Begin
        FSection  := aPatchList.Variable[ 'section' ];
        FPosition := aPatchList.Variable[ 'position'];
        aList     := aPatchList.Packet  [ 'names'   ];
        aName     := '';
        While Assigned( aList) Do
        Begin
          aName := aName + Char( aList.Variable[ 'first']);
          If aName[ 1] <> #$02
          Then aName := aName + ExtractName( aList.Packet[ 'name']);
          FNames.Add( aName);
          aList := aList.Packet[ 'next'];
        End;
        FEndOfList := 0;
      End
      Else Begin
        FSection   := 0;
        FPosition  := 0;
        FEndOfList := 1;
      End;
      FIsReply := True;
    End;

    Destructor  TPatchListMessage.Destroy; // Override;
    Begin
      FreeAndNil( FNames);
      Inherited;
    End;

    Procedure   TPatchListMessage.GetBitStream( aBitStreamList: TBitStreamList); // Override;
    Begin
      Raise EMidiMessage.Create( 'TPatchListMessage.getBitStream not implemented.', 0);
    End;

    Function    TPatchListMessage.EndOfSection ( Const aName: String): Boolean;
    Begin
      Result := Byte( aName[ 1]) = END_OF_SECTION;
    End;

    Function    TPatchListMessage.EmptyPosition( Const aName: String): Boolean;
    Begin
      Result := Byte( aName[ 1]) = EMPTY_POSITION;
    End;


{ ========
  TGetPatchListMessage = Class( TMidiMessage)
  Private
    Fpp       : Integer;
    Fssc      : Integer;
    FSection  : Integer;
    FPosition : Integer;
  Public
}

    Constructor TGetPatchListMessage.Create( aPacket: TPacket); // Overload;
    Begin
      Raise ENotImplemented.Create( 'TGetPatchListMessage.Create( TPacket)', 0);
    End;

    Constructor TGetPatchListMessage.Create( aSection, aPosition: Integer); // Overload;
    Begin
      Inherited Create;
      FCC           := $17;
      Slot          := 0;
      Fpp           := $41;
      Fssc          := $14;
      FSection      := aSection;
      FPosition     := aPosition;
      FExpectsReply := True;
    End;

    Procedure   TGetPatchListMessage.GetBitStream( aBitStreamList: TBitStreamList); // Override;
    Var
      IntStream : TIntStream;
      BitStream : TBitStream;
    Begin
      IntStream := TIntStream.Create;
      With IntStream Do
      Try
        Append( Fcc);
        Append( Slot);
        Append( Fpp);
        Append( Fssc);
        Append( FSection);
        Append( Position);
        AppendChecksum( IntStream);
        BitStream := TBitStream.Create;
        Inherited GetBitStream( IntStream, BitStream);
        aBitStreamList.Push( BitStream);
      Finally
        Free;
      End;
    End;


{ ========
  TAckMessage = Class( TMidiMessage)
  Private
    Fpid1    : Integer;
    Fpid2    : Integer;
    FAckType : Integer; // Type is a reserved word in Pascal
  Public
    Property    pid1 : Integer Read Fpid1 Write fpid1;
    Property    pid2 : Integer Read Fpid2 Write fpid2;
  Public
}

    Constructor TAckMessage.Create; // Overload;
    Begin
      Inherited Create;
      Fcc      := 0;
      Slot     := 0;
      Fpid1    := 0;
      FAckType := $7f;
      Fpid2    := 0;
      FIsReply := True;
    End;

    Constructor TAckMessage.Create( aPacket: TPacket); // Overlaod;
    Begin
      Create;
      Fcc      := aPacket.Variable[ 'cc'       ];
      Slot     := aPacket.Variable[ 'slot'     ];
      Fpid1    := aPacket.Variable[ 'data:pid1'];
      FAckType := aPacket.Variable[ 'data:type'];
      Fpid2    := aPacket.Variable[ 'data:pid2'];
    End;

    Procedure   TAckMessage.GetBitStream( aBitStreamList: TBitStreamList); // Override;
    Var
      anIntStream : TIntStream;
      aBitStream  : TBitStream;
    Begin
      anIntStream := TIntStream.Create;
      With anIntStream Do
      Try
        Append( Fcc);
        Append( Slot);
        Append( pid1);
        Append( pid2);
        Append( FAckType);
        Append( pid2);
        If FAckType = $36
        Then AppendChecksum( anIntStream);
        aBitStream := TBitStream.Create;
        GetBitStream( anIntStream, aBitStream);
        aBitStreamList.Push( aBitStream);
      Finally
        Free;
      End;
    End;


{ ========
  TGetPatchMessage = Class( TMidiMessage)
  Private
    FPid : Integer;
  Public
}

    Constructor TGetPatchMessage.Create( aSlot, aPid: Integer); // Overload;
    Begin
      Inherited Create;
      FCC  := $17;
      Slot := aSlot;
      FPid := aPid;
    End;

    Constructor TGetPatchMessage.Create( aPacket: TPacket); // Overload;
    Begin
      Raise ENotImplemented.Create( 'TGetPatchMessage.Create( TPacket)', 0);
    End;

    Procedure   TGetPatchMessage.GetBitStream( aBitStreamList: TBitStreamList); // Override;
    Var
      IntStream : TIntStream;
      BitStream : TBitStream;
    Begin
      IntStream := TIntStream.Create;
      With IntStream Do
        Try
          Append( FCC);
          Append( Slot);
          Append( FPid);
          Append( $20);
          Append( $28);
          AppendChecksum( IntStream);
          BitStream := TBitStream.Create;
          Inherited GetBitStream( IntStream, BitStream);
          aBitStreamList.Push( BitStream);

          Size := 3;
          Append( $4b);
          Append( $01);
          AppendChecksum( IntStream);
          BitStream := TBitStream.Create;
          Inherited GetBitStream( IntStream, BitStream);
          aBitStreamList.Push( BitStream);

          Size := 3;
          Append( $4b);
          Append( $00);
          AppendChecksum( IntStream);
          BitStream := TBitStream.Create;
          Inherited GetBitStream( IntStream, BitStream);
          aBitStreamList.Push( BitStream);

          Size := 3;
          Append( $53);
          Append( $01);
          AppendChecksum( IntStream);
          BitStream := TBitStream.Create;
          Inherited GetBitStream( IntStream, BitStream);
          aBitStreamList.Push( BitStream);

          Size := 3;
          Append( $53);
          Append( $00);
          AppendChecksum( IntStream);
          BitStream := TBitStream.Create;
          Inherited GetBitStream( intStream, BitStream);
          aBitStreamList.Push( BitStream);

          Size := 3;
          Append( $4c);
          Append( $01);
          AppendChecksum( IntStream);
          BitStream := TBitStream.Create;
          Inherited GetBitStream(intStream, BitStream);
          aBitStreamList.Push( BitStream);

          Size := 3;
          Append( $4c);
          Append( $00);
          AppendChecksum( IntStream);
          BitStream := TBitStream.Create;
          Inherited GetBitStream(intStream, BitStream);
          aBitStreamList.Push( BitStream);

          Size := 3;
          Append( $66);
          AppendChecksum( IntStream);
          BitStream := TBitStream.Create;
          Inherited GetBitStream(intStream, BitStream);
          aBitStreamList.Push( BitStream);

          Size := 3;
          Append( $63);
          AppendChecksum( IntStream);
          BitStream := TBitStream.Create;
          Inherited GetBitStream(intStream, BitStream);
          aBitStreamList.Push( BitStream);

          Size := 3;
          Append( $61);
          AppendChecksum( IntStream);
          BitStream := TBitStream.Create;
          Inherited GetBitStream(intStream, BitStream);
          aBitStreamList.Push( BitStream);

          Size := 3;
          Append( $4e);
          Append( $01);
          AppendChecksum( IntStream);
          BitStream := TBitStream.Create;
          Inherited GetBitStream(intStream, BitStream);
          aBitStreamList.Push( BitStream);

          Size := 3;
          Append( $4e);
          Append( $00);
          AppendChecksum( IntStream);
          BitStream := TBitStream.Create;
          Inherited GetBitStream(intStream, BitStream);
          aBitStreamList.Push( BitStream);

          Size := 3;
          Append( $68);
          AppendChecksum( IntStream);
          BitStream := TBitStream.Create;
          Inherited GetBitStream(intStream, BitStream);
          aBitStreamList.Push( BitStream);
        Finally
          Free; // IntStream
        End;
    End;


{ ========
  TPatchMessage = Class( TMidiMessage)
  Private
    FPatch       : TPatch;
    FPatchStream : TBitStream;
    FPid         : Integer;
    FChecksum    : Integer;
  Public
    Property    Patch: TPatch  Read FPatch;
    Property    Pid  : Integer Read FPid;
  Private
}

Var

  GPatchPdlFile  : String;
  GPatchProtocol : TProtocol     = Nil;
  GPatchParser   : TPacketParser = Nil;

    Class Function TPatchMessage.PatchPdlFile: String;
    Begin
      Result := GPatchPdlFile;
    End;

    Class Function TPatchMessage.PatchParser: TPacketParser;
    Begin
      Result := GPatchparser;
    End;

//  Private

    Procedure   TPatchMessage.Init;
    Begin
      If Assigned( FPatch)
      Then FreeAndNil( FPatch);
      FCC  := $1c;
      Slot := 0;
      FPid := 0;
      FExpectsReply := True;
      FIsReply      := True;
      If Not Assigned( GpatchProtocol)
      Then Begin
        GPatchProtocol := TProtocol.Create;
        GPatchProtocol.LoadPdlFile( PatchPdlFile);
        GPatchParser := GPatchProtocol.PacketParser[ 'Patch'];
      End;
    End;

    Procedure   TPatchMessage.AppendName( Const aName: String; aPatchStream: TIntStream);
    Var
      i : Integer;
    Begin
      i := 0;
      While ( i < 16) And ( i < Length( aName)) Do
      Begin
        aPatchStream.Append( Integer( aName[ i + 1]));
        Inc( i);
      End;
      If i < 16
      Then aPatchStream.Append( 0);
    End;

    Procedure   TPatchMessage.StoreEndPosition( anIntStream: TIntStream; EndPositions: TIntList);
    Var
      aPatchStream : TBitStream;
    Begin
      aPatchStream := TBitStream.Create;
      Try
        anIntStream.Reset;
        Assert( PatchParser.Generate( anIntStream, aPatchStream), 'hmm ...');
        EndPositions.Push( aPatchStream.Size Div 8 - 1);
      Finally
        aPatchStream.Free;
      End;
    End;

    Function    TPatchMessage.GetModule(
      aPatch        : TPatch;
      aSection      : TTModuleSection;
      anIndex       : Integer;
      Const aContext: String
    ): TModule;
    Begin
      If aSection = msMorph
      then Result := Nil
      Else Begin
        Result := apatch.ModuleSection[ aSection].Module[ anIndex];
        If Not Assigned( result)
        Then Raise EProtocol.CreateFmt( 'Can''t find module for %s.%d', [ aContext, anIndex]);
      End;
    End;

//  Public

    Constructor TPatchMessage.Create; // Overload;
    Begin
      Inherited Create;
      Init;
    End;

    Constructor TPatchMessage.Create( aPatch : TPatch); // Overload;
    Begin
      Create;
      FPatch := aPatch;
    End;

    Destructor  TPatchMessage.Destroy; // Override;
    Begin
      If Assigned( FPatch)
      Then FreeAndNil( FPatch);
      If Assigned( FPatchStream)
      Then FreeAndNil( FPatchStream);
      Inherited;
    End;

    Procedure   TPatchMessage.GetBitStream( aBitStreamList: TBitStreamList); // Override;
    Var
      IntStream           : TIntStream;
      SectionEndPositions : TIntList;
      sec                 : TTModuleSection;
      i                   : Integer;
      j                   : Integer;
      cnt                 : Integer;
      mph                 : TTMorph;
      Found               : Boolean;
      First               : Integer;
      Last                : Integer;
      MsgNumber           : Integer;
      PartialPatchStream  : TBitStream;
      SectionsEnded       : Integer;
      PFront              : Integer;
      NewIntStream        : TIntStream;
      NewBitStream        : TBitStream;
    Begin
      IntStream           := TIntStream.Create;
      SectionEndPositions := TIntList  .Create;

      With IntStream Do
      Try

        // Create patch bit stream

        // Name section (1)
        Append( $37); // $37 55
        AppendName( Patch.Name, IntStream);
        StoreEndPosition( IntStream, SectionEndPositions);

        // Header section (2)
        Append( $21); // $21 33
        Append( Patch.KeyboardRange[ rlMin]);
        Append( Patch.KeyboardRange[ rlMax]);
        Append( Patch.VelocityRange[ rlMin]);
        Append( Patch.VelocityRange[ rlMax]);
        Append( Patch.BendRange);
        Append( Patch.PortamentoTime);
        Append( Ord( Patch.Portamento));
        Append( 1);
        Append( Patch.RequestedVoices - 1);
        Append( 0);
        Append( Patch.SectionSeparationPos);
        Append( Patch.OctaveShift);
        Append( Ord( Patch.CableVisibility[ coRED   ]));
        Append( Ord( Patch.CableVisibility[ coBLUE  ]));
        Append( Ord( Patch.CableVisibility[ coYELLOW]));
        Append( Ord( Patch.CableVisibility[ coGRAY  ]));
        Append( Ord( Patch.CableVisibility[ coGREEN ]));
        Append( Ord( Patch.CableVisibility[ coPURPLE]));
        Append( Ord( Patch.CableVisibility[ coWHITE ]));
        Append( Ord( Patch.ModuleSection[ msCommon].VoiceRetrigger));
        Append( Ord( Patch.ModuleSection[ msPoly  ].VoiceRetrigger));
        Append( $0f); // ?
        Append( $00);
        StoreEndPosition( IntStream, SectionEndPositions);

        // Module sections (3,4)
        For sec := msPoly DownTo msCommon Do
        Begin
          Append( $4a); // $4a 74
          Append( Ord( sec));
          With Patch.ModuleSection[ sec] Do
            With Modules Do
            Begin
              IntStream.Append( Size);
              For i := 0 To Size - 1 Do
                With Items[ i] Do
                Begin
                  IntStream.Append( ModuleType);
                  IntStream.Append( Index     );
                  IntStream.Append( XPosition );
                  IntStream.Append( YPosition );
                End;
            End;
          StoreEndPosition( IntStream, SectionEndPositions);
        End;

        // Note section (5)
        Append( $69); // $69 105
        With Patch.MidiNotes Do
        Begin
          If Assigned( Items[ 0])
          Then
            With Items[ 0] Do
            Begin
              IntStream.Append( NoteNumber);
              IntStream.Append( AttackVelocity);
              IntStream.Append( ReleaseVelocity);
            End
            Else Begin
              IntStream.Append( 0);
              IntStream.Append( 0);
              IntStream.Append( 0);
            End;
          Append( Size - 2);
          For i := 1 To Size - 1 Do
            With Items[ i] Do
            Begin
              IntStream.Append( NoteNumber);
              IntStream.Append( AttackVelocity);
              IntStream.Append( ReleaseVelocity);
            End;
        End;
        StoreEndPosition( IntStream, SectionEndPositions);

        // Cable sections (6,7)
        For sec := msPoly DownTo msCommon Do
        Begin
          Append( $52); // $52 82
          Append( Ord( sec));
          With Patch.ModuleSection[ sec] Do
            With Cables Do
            Begin
              IntStream.Append( Size);
              For i := 0 To Size - 1 Do
                With Cables[ i] Do
                Begin
                  IntStream.Append( Ord( Color));
                  IntStream.Append( SourceModule.Index);
                  IntStream.Append( SourceConnector);
                  IntStream.Append( Ord( SourceConnectorType));
                  IntStream.Append( DestinationModule.Index);
                  IntStream.Append( DestinationConnector);
                  // IntStream.Append( Ord( DestinationConnectorType));
                End;
            End;
          StoreEndPosition( IntStream, SectionEndPositions);
        End;

        // Parameter sections (8,9)
        For sec := msPoly DownTo msCommon Do
        Begin
          Append( $4d); // $4d 77
          Append( Ord( sec));
          With Patch.ModuleSection[ sec] Do
          Begin
            // Count modules having parameters
            cnt := 0;
            With Modules Do
              For i := 0 To Size - 1 Do
                With Items[ i] Do
                  If ParameterCount > 0
                  Then Inc( cnt);
            IntStream.Append( cnt);

            // Append modules having parameters
            With Modules Do
              For i := 0 To Size - 1 Do
                With Items[ i] Do
                  If ParameterCount > 0
                  Then Begin
                    IntStream.Append( Index     );
                    IntStream.Append( ModuleType);
                    For j := 0 To ParameterCount - 1 Do
                    Begin
                      IntStream.Append( Parameter[ j]);
                    End;
                  End;
          End;
          StoreEndPosition( IntStream, SectionEndPositions);
        End;

        // Morph section (10)
        Append( $65); // $65 101
        cnt := 0;
        For mph := mtMorph1 To mtMorph4 Do
          With Patch.Morph[ mph] Do
          Begin
            IntStream.Append( Value);
            Inc( cnt, MorphMaps.Size);
          End;
        //
        For mph := mtMorph1 To mtMorph4 Do
          With Patch.Morph[ mph] Do
            IntStream.Append( Ord( KeyBoardAssignment));
        //
        Append( cnt);
        //
        For mph := mtMorph1 To mtMorph4 Do
          With Patch.Morph[ mph].MorphMaps Do
            For i := 0 To Size - 1 Do
              With Items[ i] Do
              Begin
                IntStream.Append( Ord( Section));
                IntStream.Append( Ord( Index));
                IntStream.Append( Parameter);
                IntStream.Append( Ord( mph));
                IntStream.Append( Range);
              End;
        //
        StoreEndPosition( IntStream, SectionEndPositions);

        // Knob section (11)
        Append( $62); // $62 98
        For i := 0 To 22 Do
        Begin
          Found := False;
          With Patch.KnobMaps Do
          Begin
            For j := 0 To Size - 1 Do
            Begin
              If Items[ j].Knob = i
              Then Begin
                Found := True;
                IntStream.Append( 1);
                IntStream.Append( Ord( Items[ j].ModuleSection));
                If Items[ j].ModuleSection = msMorph
                Then IntStream.Append( 1)
                Else IntStream.Append( Items[ j].Module.Index);
                IntStream.Append( Items[ j].Parameter);
              End;
            End;
            If Not Found
            Then IntStream.Append( 0);
          End;
        End;
        StoreEndPosition( IntStream, SectionEndPositions);

        // Control section (12)
        Append( $60); // $60 96
        With Patch.CtrlMaps Do
        Begin
          IntStream.Append( Size);
          For i := 0 To Size - 1 Do
            With Items[ i] Do
            Begin
              IntStream.Append( CC);
              IntStream.Append( Ord( ModuleSection));
              If ModuleSection = msMorph
              Then IntStream.Append( 1)
              Else IntStream.Append( Module.Index);
              IntStream.Append( Parameter);
            End;
        End;
        StoreEndPosition( IntStream, SectionEndPositions);

        // Custom sections (13,14)
        For sec := msPoly DownTo msCommon Do
        Begin
          Append( $5b); // $5b 91
          Append( Ord( sec));
          With Patch.ModuleSection[ sec] Do
          Begin
            cnt := 0;
            With Modules Do
              For i := 0 To Size - 1 Do
                With Items[ i] Do
                  If CustomCount > 0
                  Then Inc( cnt);
          End;
          Append( cnt);
          With Patch.ModuleSection[ sec] Do
            With Modules Do
              For i := 0 To Size - 1 Do
                With Items[ i] Do
                  If CustomCount > 0
                  Then Begin
                    IntStream.Append( Index);
                    IntStream.Append( CustomCount);
                    For j := 0 To CustomCount - 1 Do
                      IntStream.Append( Custom[ j]);
                  End;

          StoreEndPosition( IntStream, SectionEndPositions);
        End;

        // Module name section (15,16)
        For sec := msPoly DownTo msCommon Do
        Begin
          Append( $5a); // $5a 90
          Append( Ord( sec));
          With Patch.ModuleSection[ sec] Do
            With Modules Do
            Begin
              IntStream.Append( Size);
              For i := 0 To Size - 1 Do
                With Items[ i] Do
                Begin
                  IntStream.Append( Index);
                  AppendName( Name, IntStream);
                End;
            End;
          StoreEndPosition( IntStream, SectionEndPositions);
        End;

        If Assigned( FPatchStream)
        Then FreeAndNil( FPatchStream);
        FPatchStream := TBitStream.Create;
        IntStream.Reset;
        PatchParser.Generate( IntStream, FPatchStream);

        // Create sysex messages

        First     := 1;
        Last      := 0;
        MsgNumber := 0;
        PFront    := 0;

        While FPatchStream.HasAvailable( 8) Do
        Begin
          // Get data for one sysex packet
          PartialPatchStream := TBitStream.Create;
          SectionsEnded := 0;
          cnt           := 0;
          While
            FPatchStream.HasAvailable( 8) And
            ( cnt < 166)
          Do
          Begin
            PartialPatchStream.Append( FPatchStream.GetInt( 8), 8);
            If cnt = SectionEndPositions.Items[ PFront] - 166 * MsgNumber
            Then Begin
              Inc( SectionsEnded);
              Inc( PFront);        // SectionEndPositions.pop_front()
            End;
            Inc( cnt);
          End;
          Inc( MsgNumber);

          If Not FPatchStream.HasAvailable( 8)
          Then Last := 1;

          // Pad, extra bits are ignored later.
          PartialPatchStream.Append( 0, 6);

          // Generate sysex bistream
          NewIntStream := TIntStream.Create;
          Try
            NewIntStream.Append( FCC + First + 2 * Last);
            First := 0;
            NewIntStream.Append( Slot);
            NewIntStream.Append( $01);
            NewIntStream.Append( SectionsEnded);
            While PartialPatchStream.HasAvailable( 7) Do
              NewIntStream.Append( PartialPatchStream.GetInt( 7));
            AppendChecksum( NewIntStream);

            // Generate sysex bitstream
            NewBitStream := TBitStream.Create;
            GetBitStream( NewIntStream, NewBitStream);
            aBitStreamList.Push( NewBitStream);
          Finally
            FreeAndNil( NewIntStream);
          End;
        End;

      Finally
        FreeAndNil( IntStream);
        FreeAndNil( SectionEndPositions);
      End;
    End;

    Procedure   TPatchMessage.Append( aPacket: TPacket);
    Begin
      Slot    := aPacket.Variable[ 'Slot'     ];
      FPid    := aPacket.Variable[ 'data:pid' ];
      aPacket := aPacket.Packet  [ 'data:next'];
      While Assigned( aPacket) Do
      Begin
        FPatchStream.Append( aPacket.Variable[ 'data'], 7);
        aPacket := aPacket.Packet[ 'next'];
      End;
      With FPatchStream Do
      Begin
        Size := Size - 7;          // Remove checksum
        Size := ( Size Div 8) * 8; // Remove padding
      End;
    End;

    Procedure   TPatchMessage.GetPatch( aPatch: TPatch);
    Var
      Packet         : TPacket;
      Section        : Tpacket;
      SectionData    : TPacket;
      aModuleSection : TModuleSection;
      aPacketList    : TPacketList;
      bPacketList    : TPacketList;
      i              : Integer;
      j              : Integer;
      aModule        : TModule;
      aSection       : TTModuleSection;
      aMorph         : TMorph;
      aMorphMap      : TMorphMap;
      TmpPacket      : TPacket;
    Begin
      Packet := TPacket.Create;
      Try
        FPatchStream.Reset;
        If PatchParser.Parse( FPatchStream, Packet)
        Then
          While Assigned( Packet) Do
          Begin
            Section     := Packet.Packet[ 'section'];
            SectionData := packet.Packet[ 'data'   ];
            Case Section.Variable[ 'type'] Of

              $37, $27 : // 55, 39 Name section
                Begin
                  aPatch.Name := ExtractName( SectionData.Packet[ 'name']);
                End;

              $21 : // 33 Header section
                Begin
                  With SectionData Do
                  Begin
                    aPatch.KeyboardRange[ rlMIN] := Variable[ 'krangemin'];
                    aPatch.KeyboardRange[ rlMAX] := Variable[ 'krangemax'];
                    aPatch.VelocityRange[ rlMIN] := Variable[ 'vrangemin'];
                    aPatch.VelocityRange[ rlMAX] := Variable[ 'vrangemax'];
                    aPatch.BendRange             := Variable[ 'brange'   ];
                    aPatch.PortamentoTime        := Variable[ 'ptime'    ];
                    aPatch.Portamento            := TPortamento( SectionData.Variable[ 'portamento']);
                    aPatch.RequestedVoices       := SectionData.Variable[ 'voices'];
                    aPatch.SectionSeparationPos  := SectionData.Variable[ 'sspos' ];
                    aPatch.OctaveShift           := SectionData.Variable[ 'octave'];
                    aPatch.CableVisibility[ coRED   ] := Boolean( SectionData.Variable[ 'red'   ]);
                    aPatch.CableVisibility[ coBLUE  ] := Boolean( SectionData.Variable[ 'blue'  ]);
                    aPatch.CableVisibility[ coYELLOW] := Boolean( SectionData.Variable[ 'yellow']);
                    aPatch.CableVisibility[ coGRAY  ] := Boolean( SectionData.Variable[ 'gray'  ]);
                    aPatch.CableVisibility[ coGREEN ] := Boolean( SectionData.Variable[ 'green' ]);
                    aPatch.CableVisibility[ coPURPLE] := Boolean( SectionData.Variable[ 'purple']);
                    aPatch.CableVisibility[ coWHITE ] := Boolean( SectionData.Variable[ 'white' ]);
                    aPatch.ModuleSection[ msCOMMON].VoiceRetrigger := TVoiceRetrigger( SectionData.Variable[ 'cretrigger']);
                    aPatch.ModuleSection[ msPOLY  ].VoiceRetrigger := TVoiceRetrigger( SectionData.Variable[ 'pretrigger']);
                  End;
                End;

              $4a : // 74 Module section
                Begin
                  aModuleSection := aPatch.ModuleSection[ TTModuleSection( SectionData.Variable  [ 'section'])];
                  aPacketList    := SectionData.PacketList[ 'modules'];
                  With aPacketList Do
                    For i := 0 To Size - 1 Do
                      With Items[ i] Do
                      Begin
                        aModule := aModuleSection.NewModule( Variable[ 'type']);
                        aModule.SetPosition( Variable[ 'xpos'], Variable[ 'ypos']);
                      End;
                End;

              $69 : // 105 Note section
                Begin
                  With SectionData.Packet[ 'note1'] Do
                    Patch.NewMidiNote( Variable[ 'value'], Variable[ 'attack'], Variable[ 'release']);
                  With SectionData.Packet[ 'note2'] Do
                    Patch.NewMidiNote( Variable[ 'value'], Variable[ 'attack'], Variable[ 'release']);
                  With SectionData.PacketList[ 'notes'] Do
                    For i := 0 To Size - 1 Do
                      With Items[ i] Do
                        Patch.NewMidiNote( Variable[ 'value'], Variable[ 'attack'], Variable[ 'release']);
                End;

              $52 : // 82 Cable section
                Begin
                  aModuleSection := aPatch.ModuleSection[ TTModuleSection( SectionData.Variable  [ 'section'])];
                  aPacketList    := SectionData.PacketList[ 'cables'];
                  With aPacketList Do
                    For i := 0 To Size - 1 Do
                      With Items[ i] Do
                        aModuleSection.NewCable(
                          TTColor( Variable[ 'color']),
                          Variable[ 'destination'],
                          Variable[ 'input'],
                          ctInput,
                          Variable[ 'source'],
                          Variable[ 'inputOutput'],
                          TTConnector( Variable[ 'type'])
                        );
                End;

              $4d : // 77 Parameter section
                Begin
                  aSection    := TTModuleSection( SectionData.Variable[ 'section']);
                  aPacketList := SectionData.PacketList[ 'parameters'];
                  With aPacketList Do
                    For i := 0 To Size - 1 Do
                      With Items[ i] Do
                      Begin
                        aModule := GetModule( aPatch, aSection, Variable[ 'index'], 'parameter');
                        With Packet[ 'parameters'].AllVariables Do
                          For j := 0 To Size - 1 Do
                            aModule.Parameter[ j] := Items[ j];
                      End;
                End;

              $65 : // 101 MorphMap section
                Begin
                  aPatch.Morph[ mtMORPH1].Value := SectionData.Variable[ 'morph1'];
                  aPatch.Morph[ mtMORPH2].Value := SectionData.Variable[ 'morph2'];
                  aPatch.Morph[ mtMORPH3].Value := SectionData.Variable[ 'morph3'];
                  aPatch.Morph[ mtMORPH4].Value := SectionData.Variable[ 'morph4'];

                  aPatch.Morph[ mtMORPH1].KeyboardAssignment :=
                    TKeyboardAssignment( SectionData.Variable[ 'keyboard1']);
                  aPatch.Morph[ mtMORPH2].KeyboardAssignment :=
                    TKeyboardAssignment( SectionData.Variable[ 'keyboard2']);
                  aPatch.Morph[ mtMORPH3].KeyboardAssignment :=
                    TKeyboardAssignment( SectionData.Variable[ 'keyboard3']);
                  aPatch.Morph[ mtMORPH4].KeyboardAssignment :=
                    TKeyboardAssignment( SectionData.Variable[ 'keyboard4']);

                  aPacketList := SectionData.PacketList[ 'morphs'];
                  With aPacketList Do
                    For i := 0 To Size - 1 Do
                      With Items[ i] Do
                      Begin
                        aMorph    := aPatch.Morph[ TTMorph( Variable[ 'morph'])];
                        aModule   := GetModule( aPatch, TTmoduleSection( Variable[ 'section']), Variable[ 'module'], 'morph');
                        aMorphMap := aMorph.NewMorphmap( TTModuleSection( Variable[ 'section']), aModule, Variable[ 'parameter']);
                        j         := Variable[ 'range'];
                        With aMorphMap Do
                          If j > 127
                          Then Range := j - 256
                          Else Range := j;
                      End;
                End;

              $62 : // 98 KnobMap section
                Begin
                  For i := 0 To 22 Do
                  Begin
                    With Section.Packet[ Format( 'knob%d', [ i])] Do
                      If Boolean( Variable[ 'assigned'])
                      Then Begin
                        TmpPacket := PacketList[ 'assignment'].Items[ 0];
                        aSection  := TTModuleSection( Variable[ 'section']);
                        aModule   := GetModule( aPatch, aSection, TmpPacket.Variable[ 'module'], 'knob');
                        aPatch.NewKnobMap( aSection, aModule, TmpPacket.Variable[ 'parameter']).Knob := i;
                      End;
                  End;
                End;

              $60 : // 96 ControlMap section
                Begin
                  aPacketList := SectionData.PacketList[ 'controls'];
                  With aPacketList Do
                    For i := 0 To Size - 1 Do
                      With Items[ i] Do
                      Begin
                        aSection := TTModuleSection( Variable[ 'section']);
                        aModule := GetModule( aPatch, aSection, Variable[ 'module'], 'control');
                        aPatch.NewCtrlMap( aSection, aModule, Variable[ 'parameter']).CC := Variable[ 'control'];
                      End;
                End;

              $5b : // 91 Custom Section
                Begin
                  aSection    := TTModuleSection( SectionData.Variable[ 'section']);
                  aPacketList := SectionData.PacketList[ 'customModules'];
                  With aPacketList Do
                    For i := 0 To Size - 1 Do
                      With Items[ i] Do
                      Begin
                        aModule     := GetModule( aPatch, aSection, Variable[ 'index'], 'custom');
                        bPacketList := PacketList[ 'customValues'];
                        With bPAcketList Do
                          For j := 0 To Size - 1 Do
                            With Items[ i] Do
                              aModule.Custom[ j] := Variable[ 'value'];
                      End;
                End;

              $5a : // 90 NameDump section
                Begin
                  If Assigned( SectionData)
                  Then Begin
                    aSection    := TTmoduleSection( SectionData.Variable[ 'section']);
                    aPacketList := SectionData.PacketList[ 'moduleNames'];
                    With aPacketList Do
                      For i := 0 To Size - 1 Do
                        With Items[ i] Do
                        Begin
                          aModule := GetModule( aPatch, aSection, Variable[ 'index'], 'name');
                          aModule.Name := ExtractName( Packet[ 'name']);
                        End;
                  End;
                End;

            End; // Case Section.Variable[ 'type'] Of
            Packet := Packet.Packet[ 'next'];
          End // While Assigned( Packet)
        Else // If PatchParser.Parse( FPatchStream, Packet)
          Raise EPatch.Create( 'Could not read patch.');
      Finally
        FreeAndNil( Packet); // @@@@ sure about this ??
      End;
    End;

//  Public

    Class Procedure TPatchMessage.UsePdlFile( Const aFileName: String; aTracer: TTracer = Nil);
    Begin
      GPatchPdlFile := aFileName;
      If Assigned( GPatchProtocol)
      Then FreeAndNil( GPatchProtocol);
      GPatchProtocol := TProtocol.Create;
      GPatchProtocol.LoadPdlFile( GPatchPdlFile);
      GPatchParser := GPatchProtocol.PacketParser[ 'Patch'];
      GPatchProtocol.UseTracer( aTracer);
    End;


{ ========
  TEnqueuedPacket = Class
  Private
    FContent      : TByteList;
    FExpectsReply : Boolean;
  Public
    Property    Content     : TByteList Read FContent;
    Property    ExpectsReply: Boolean   Read FExpectsReply;
  Public
}

    Constructor TEnqueuedPacket.Create( aContent: TByteList; anExpectsReply: Boolean);
    Begin
      Inherited Create;
      FContent      := aContent;
      FExpectsReply := anExpectsReply;
    End;

    Destructor  TEnqueuedPacket.Destroy; // Override;
    Begin
      If Assigned( FContent)
      Then FreeAndNil( FContent);
      Inherited;
    End;


{ ========
  TNMProtocolDriver = Class
  Private
    FOnDriverData    : TOnDriverData;               // Transmit to driver
    FOnMidiMessage   : TOnMidiMessage;              // Application midi message handler
    FOnUnparsable    : TOnUnparsable;
    FReceiveState    : ( rsIdle, rsInSysex);
    FRxBuffer        : TByteList;
    FTxBuffer        : TPacketQueue;
    FTimeOutTime     : DWORD;                       // ms, 0 for no timeout
    FTimeoutInterval : DWORD;
  Protected
    Property    TimeoutTime: DWORD Read FTimeoutTime Write FTimeoutTime;
  Public
    Property    OnDriverData   : TOnDriverData  Read FOnDriverData    Write FOnDriverData;
    Property    OnMidiMessage  : TOnMidiMessage Read FOnMidiMessage   Write FOnMidiMessage;
    Property    OnUnparsable   : TOnUnparsable  Read FOnUnparsable    Write FOnUnparsable;
    Property    TimeoutInterval: DWORD          Read FTimeoutInterval Write FTimeoutInterval;
  Private
}

    Procedure   TNMProtocolDriver.NotifyListeners( aMessage: TMidiMessage);
    Begin
      If Assigned( FOnMidiMessage)
      Then FOnMidiMessage( Self, aMessage);
    End;

    Procedure   TNMProtocolDriver.NotifyUnparsable( aData: TByteList);
    Begin
      If Assigned( FOnUnparsable)
      Then FOnUnparsable( Self, aData);
    End;

    Function    TNMProtocolDriver.ReplyExpected: Boolean;
    Begin
      Result := TimeoutTime > 0;
    End;

    Function    TNMProtocolDriver.TimeoutOccured: Boolean;
    Begin
      Result := ReplyExpected And ( TimeGetTime > TimeoutTime);
    End;

    Procedure   TNMProtocolDriver.StartTimeout( aTimeInms: DWORD);
    Begin
      TimeoutTime := TimeGetTime + aTimeInms;
    End;

    Procedure   TNMProtocolDriver.StopTimeout;
    Begin
      TimeOutTime := 0;
    End;

//  Protected

    Procedure   TNMProtocolDriver.AcceptMidiMessage( aMidiMsg: TMidiMessage);
    Begin
      NotifyListeners( aMidiMsg);
      If ReplyExpected
      Then Begin
        If aMidiMsg.IsReply // Is it a reply ?
        Then Begin
          FTxBuffer.Drop;
          Stoptimeout;
          DoTransmit;
        End;
        // Else : not a reply expected, go on waiting
      End
      Else DoTransmit;
    End;

    Procedure   TNMProtocolDriver.AcceptMidiPacket( aData: TByteList);
    Var
      i        : Integer;
      BitStream: TBitStream;
      MidiMsg  : TMidiMessage;
    Begin
      If Assigned( aData) And ( aData.Size > 0)
      Then Begin
        BitStream := TBitStream.Create;
        Try
          For i := 0 To aData.Size - 1 Do
            BitStream.Append( aData.Items[ i], 8);
          MidiMsg := TMidiMessage.CreateMsg( BitStream);
          If Assigned( MidiMsg)
          Then
            Try
              AcceptMidiMessage( MidiMsg);
            Finally
              MidiMsg.Free;
            End
          Else Begin
            NotifyUnparsable( aData);
            // Could not parse reply, retransmit
            NotifyUnparsable( aData);
            DoTransmit;
          End;
        Finally
          BitStream.Free;
        End;
      End;
    End;

  {$Q-,R-}

    Procedure   TNMProtocolDriver.DoTransmit;
    Var
      aMsg : TEnqueuedPacket;
    Begin
      aMsg := FTxBuffer.Peek;
      If Assigned( aMsg)
      Then Begin
        If Assigned( FOnDriverData)
        Then FOnDriverData( Self, aMsg.Content);
        If aMsg.ExpectsReply And Assigned( FOnDriverData)
        Then StartTimeout( TimeoutInterval)
        Else Begin
          StopTimeout;    // no timeout when no reply is expected
          FTxBuffer.Drop  // Don't remember packets that need no reply
        End;
      End
      Else StopTimeout;
    End;

  {$Q+,R+}

//  Public


    Constructor TNMProtocolDriver.Create;
    Begin
      Inherited Create;
      FRxBuffer        := TByteList.Create;
      FTxBuffer        := TPacketQueue.Create;
      FTimeoutInterval := 3500; // ms
      StopTimeout;
    End;

    Destructor  TNMProtocolDriver.Destroy; // Override;
    Begin
      If Assigned( FRxBuffer)
      Then FreeAndNil( FRxBuffer);
      Inherited;
      If Assigned( FTXBuffer)
      Then FreeAndNil( FTxBuffer);
    End;

    Procedure   TNMProtocolDriver.ClearTransmitter;
    Begin
      FTxBuffer.Clear;
    End;
    
    Procedure   TNMProtocolDriver.HeartBeat;
    Begin
      If TimeoutOccured
      Then Begin
        // FTxBuffer.Drop;
        DoTransmit;
      End;
    End;

    Procedure   TNMProtocolDriver.AcceptDriverData( aData: TByteList); // Accept driver data
    Var
      i : Integer;
    Begin
      If Assigned( aData)
      Then
        With aData Do
        Begin
          For i := 0 To Size - 1 Do
            Case FReceiveState Of
              rsIdle :
                Begin
                  If Items[ i] = SYSEX_START
                  Then Begin
                    FRxBuffer.Clear;
                    FRxBuffer.Push( SYSEX_START);
                    FReceiveState := rsInSysex;
                  End;
                End;
              rsInSysex :
                Begin
                  FRxBuffer.Push( Items[ i]);
                  If Items[ i] = SYSEX_END
                  Then Begin
                    AcceptMidiPacket( FRxBuffer);
                    FReceiveState := rsIdle;
                  End;
                End;
            End;
          Clear;
        End;
    End;

    Procedure   TNMProtocolDriver.AcceptRawDriverData( aData: PBytes; aSize: Integer);
    Var
      aByteList : TByteList;
      i         : Integer;
    Begin
      aByteList := TByteList.Create;
      With aByteList Do
        Try
          Size := aSize;
          For i := 0 To aSize - 1 Do
            Items[ i] := aData^[ i];
          AcceptDriverData( aByteList);
        Finally
          Free;
        End;
    End;

    Procedure   TNMProtocolDriver.Send( aMessage: TMidiMessage);
    Var
      BitStreamList : TBitStreamList;
      TxPacket      : TByteList;
      i             : Integer;
    Begin
      If Assigned( FOnDriverData)
      Then Begin
        BitStreamList := TBitStreamList.Create;
        aMessage.GetBitStream( BitStreamList);
        With BitStreamList Do
          For i := 0 To Size - 1 Do
          Begin
            TxPacket := TByteList.Create;
            With Items[ i] Do
              While HasAvailable( 8) Do
                TxPacket.Push( GetInt( 8));
            FTxBuffer.Push(
              TEnqueuedPacket.
                Create(
                  TxPacket,
                  aMessage.ExpectsReply
                )
            );
            If ( FTxBuffer.Count = 1)
            Then DoTransmit;
          End;
      End;
    End;

Initialization

  TMidiMessage.SetPdlFile( ApplicationPath + PathDelim + 'midi.pdl');

Finalization

  If Assigned( GTMidiMessageProtocol)
  Then FreeAndNil( GTMidiMessageProtocol);

  If Assigned( GPatchProtocol)
  Then FreeAndNil( GPatchProtocol);

End.

