Unit LexLib;

// Standard Lex library unit for TP Lex Version 3.0. 2-11-91 AG


//  ////////////////////////////////////////////////////////////////////////////
//
//  Modifications Copyright (C) 2003 Jan Punter
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//  ////////////////////////////////////////////////////////////////////////////
//
// Modified to be used under Delphi 7 by Jan Punter. 2003-10-13
//
// - added support for include files
// - changed short strings into long strings
// - I/O runs through stringlists rather than through standard I/O.
// - made yyFileName, yyLineNo, yyColNo work to point to syntax errors.
//
// 2003-11-22, made yysLength integer instead of byte.
//
//  ////////////////////////////////////////////////////////////////////////////


Interface

{
   The Lex library unit supplies a collection of variables and routines
   needed by the lexical analyzer routine yylex and application programs
   using Lex-generated lexical analyzers. It also provides access to the
   input/output streams used by the lexical analyzer and the text of the
   matched string, and provides some utility functions which may be used
   in actions.

   This `standard' version of the LexLib unit is used to implement lexical
   analyzers which read from and write to MS-DOS files (using standard input
   and output, by default). It is suitable for many standard applications
   for lexical analyzers, such as text conversion tools or compilers.

   However, you may create your own version of the LexLib unit, tailored to
   your target applications. In particular, you may wish to provide another
	 set of I/O functions, e.g., if you want to read from or write to memory
   instead to files, or want to use different file types.

  Variables:

   The variable yytext contains the current match, yyleng its length.
   The variable yyline contains the current input line, and yylineno and
   yycolno denote the current input position (line, column). These values
   are often used in giving error diagnostics (however, they will only be
   meaningful if there is no rescanning across line ends).

   The variables yyinput and yyoutput are the text files which are used
   by the lexical analyzer. By default, they are assigned to standard
   input and output, but you may change these assignments to fit your
	 target application (use the Turbo Pascal standard routines assign,
   reset, and rewrite for this purpose).
}


// Disable some .net related warnings.

{$WARN UNSAFE_TYPE OFF}
{$WARN UNSAFE_CODE OFF}
{$WARN UNSAFE_CAST OFF}


Uses

	SysUtils, Classes, Contnrs;

Type

  ELex = Class( Exception);

Var

  yyinput_text : String;               // Input buffer
  yyoutput     : TStringList;          // Output strings
  yyline       : String;          // current input line
  // yyline       : ShortString;          // current input line
  yyFileName   : String;               // Current file name, if any
  yyposition   : Integer;              // Currentcharacter position
  yylineno     : Integer;              // Current line nr
  yycolno      : Integer;              // current input position
  yytext       : String;          // matched text (should be considered r/o)
  // yyleng       : Byte Absolute yytext; // length of matched text
  yydebug      : Boolean;              // set to true to enable debugging output of parser
  yyErrorCount : Integer;              // Number of syntax errors detected


{

  I/O routines:

   The routines:

     get_char
     unget_char
     put_char

   are used to implement access to the input and output files. Since \n
   (newline) for Lex means line end, the I/O routines have to translate MS-DOS
   line ends (carriage-return/line-feed) into newline characters and vice versa.
   Input is buffered to allow rescanning text (via unput_char).

   The input buffer holds the text of the line to be scanned. When the input
	 buffer empties, a new line is obtained from the input stream. Characters
   can be returned to the input buffer by calls to unget_char. At end-of-file
   a null character is returned.

   The input routines also keep track of the input position and set the
   yyline, yylineno, yycolno variables accordingly.

   Since the rest of the Lex library only depends on these three routines
   (there are no direct references to the yyinput and yyoutput files or
	 to the input buffer), you can easily replace get_char, unget_char and
   put_char by another suitable set of routines, e.g. if you want to read
   from/write to memory, etc.
}

  Function yyleng: Integer;
  {
    replaces yyleng variable, yytext must be a long string.
  }


  Function get_char: Char;
  {
    obtain one character from the input file (null character at end-of-file).
  }

  Procedure unget_char( c: Char );
  {
    return one character to the input file to be reread in subsequent calls to
    get_char.
  }

  Procedure put_char( c: Char );
  {
    write one character to the output file.
  }

  Procedure yyStringOut( Const S: String);
  Procedure yyStringOutFmt( Const aFormat: String; Const anArgs: Array Of Const);
  {
    As put_char, but for strings
  }

// Utility routines:

  Procedure echo;
  {
    echoes the current match to the output stream.
  }

  Procedure yymore;
  {
    append the next match to the current one.
  }

  Procedure yyless( n: Integer);
  {
    truncate yytext to size n and return the remaining characters to the input
    stream.
  }

  Procedure reject;
  {
    reject the current match and execute the next one.

    reject does not actually cause the input to be rescanned; instead, internal
    state information is used to find the next match. Hence you should not try
    to modify the input stream or the yytext variable when rejecting a match.
  }

  Procedure return ( n: Integer);
  Procedure returnc( c: Char);
  {
    sets the return value of yylex.
  }

  Procedure start( state: Integer);
  {
    puts the lexical analyzer in the given start state; state=0 denotes the
    default start state, other values are user-defined.
  }

{ yywrap:

     The yywrap function is called by yylex at end-of-file ( unless you have
     specified a rule matching end-of-file). You may redefine this routine
     in your Lex program to do application-dependent processing at end of
     file. In particular, yywrap may arrange for more input and return false
     in which case the yylex routine resumes lexical analysis.
}

  Function yywrap : Boolean;
  {
    The default yywrap routine supplied here closes input and output files and
    returns true ( causing yylex to terminate).
  }


  Procedure yyInclude( Const aFileName: String);
  {
    Push the current input state and start interpreting the file named
    aFileName.
  }

  Function  Printable( anInt: Integer): String;
  Procedure yyDebugOut( Const S: String);
  Procedure yyDebugOutFmt( Const aFormat: String; Const anArgs: Array Of Const);
  {
    When yydebug is set True send utput to the output stream
  }

// The following are the internal data structures and routines used by the
// lexical analyzer routine yylex; they should not be used directly.

Var

  yystate    : Integer; // current state of lexical analyzer
  yyactchar  : Char;    // current character
  yylastchar : Char;    // last matched character (#0 if none)
  yyrule     : Integer; // matched rule
  yyreject   : Boolean; // current match rejected?
  yydone     : Boolean; // yylex return value set?
  yyretval   : Integer; // yylex return value

  Procedure yynew;
  {
    starts next match; initializes state information of the lexical analyzer
  }

  Procedure yyscan;
  {
    gets next character from the input stream and updates yytext and yyactchar
    accordingly
  }

  Procedure yymark( n: Integer);
  {
    marks position for rule no. n
  }

  Procedure yymatch( n: Integer);
  {
    declares a match for rule number n
  }

  Function yyfind( Var n: Integer): Boolean;
  {
    finds the last match and the corresponding marked position and adjusts the
    matched string accordingly;

    returns:

      True if a rule has been matched, False otherwise.
      n : the number of the matched rule.
  }

  Function yydefault: Boolean;
  {
    executes the default action (copy character); returns true unless at
    end-of-file
  }

  Procedure yyclear;
  {
    reinitializes state information after lexical analysis has been finished
  }

  Procedure yyinit;


Implementation


{$WARN UNSAFE_TYPE OFF}



  Procedure fatal( aMsg: String);
	Begin
    Inc( yyErrorCount);
		Raise ELex.CreateFmt( 'FATAL ERROR : ''%s''', [ aMsg]);
	End;

Const

  nl        = #10;  // newline character
  max_chars = 16384;

Var

  eoi					: Integer;
  bufptr      : Integer;
  ungetbuf    : array [ 0 .. max_chars] of Char;
  ungetbufptr : Integer;


Type

  TInputState = Class( TObject)
  Private
    FFileName : String;
    FData     : String;
    FPosition : Integer;
    FLineNo   : Integer;
    FColNo    : Integer;
    FBufPtr   : Integer;
  Public
    Constructor Create( Const aFileName, aData: String; aPosition, aLineNr, aColNr: Integer; aBufPtr: Integer);
  End;

  TFileStack = Class( TStack)
  Public
    Procedure   Push( Const aFileName: String);
    Procedure   Pop;
    Destructor  Destroy;                                               Override;
  End;

Var

  FileStack  : TFileStack;


{ ========
  TInputState = Class( TObject)
  Private
    FFileName : String;
    FData     : String;
    FPosition : Integer;
    FLineNo   : Integer;
    FColNo    : Integer;
    FBufPtr   : Integer;
  Public
}

    Constructor TInputState.Create( Const aFileName, aData: String; aPosition, aLineNr, aColNr: Integer; aBufPtr: Integer);
    Begin
      Inherited Create;
      FFileName := aFileName;
      FData     := aData;
      FPosition := aPosition;
      FLineNo   := aLineNr;
      FColNo    := aColNr;
      FBufPtr   := aBufPtr;
    End;


{ ========
  TFileStack = Class( TStack)
  Public
}

    Procedure   TFileStack.Push( Const aFileName: String);
    Var
      aFile : File Of Byte;
    Begin
      If FileExists( aFileName)
      Then Begin
        PushItem( TInputState.Create( yyFileName, yyinput_text, yyposition, yylineno, yycolno, bufptr));
        yyFileName := aFileName;
        AssignFile( aFile, yyFileName);
        Reset( aFile);
        Try
          SetLength( yyinput_text, FileSize( aFile));
          BlockRead( aFile, PChar( yyinput_text)^, Length( yyinput_text));
          yylineno   := 1;
          yycolno    := 0;
          bufptr     := 0;
          yyposition := 0;
          eoi        := Length( yyinput_text);
        Finally
          CloseFile( aFile);
        End;
      End;
    End;

    Procedure   TFileStack.Pop;
    Var
      anItem : TInputState;
    Begin
      If Count > 0
      Then Begin
        anItem := TInputState( PopItem);
        With anItem Do
        Begin
          yyFileName   := FFileName;
          yyinput_text := FData;
          yyPosition   := FPosition;
          yylineno     := FLineNo;
          yycolno      := FColNo;
          bufptr       := FBufPtr;
          eoi          := Length( yyinput_text);
        End;
      End
      Else Begin
        yyinput_text := '';
      End;
    End;

    Destructor  TFileStack.Destroy; // Override;
    Var
      anItem : TInputState;
    Begin
      While Count > 0 Do
      Begin
        anItem := TInputState( PopItem);
        anItem.Free;
      End;
      Inherited;
    End;


// I/O routines :

  Function yyleng: Integer;
  Begin
    Result := Length( yyText);
  End;

  Function get_char: Char;
	Begin
		Inc( yycolno);
    Inc( yyposition);
		If ungetbufptr > 0
    Then Begin
			Dec( ungetbufptr);
			Result := ungetbuf[ ungetbufptr];
			exit;
		End;

		If bufptr = 0
    Then Begin
			eoi    := Length( yyinput_text);
			bufptr := 1;
		End;

		If bufptr <= eoi
    Then Begin
			Result := yyinput_text[ bufptr];
			Inc( bufptr);
			// If Result = #13
      // Then Result := ' ';
		End
    Else Result := #0; 

    If Result = #10
    Then Begin
      inc( yylineno);
      yycolno := 0;
    End;
	End;


  Procedure unget_char( c: Char);
	Begin
		If bufptr = max_chars
    Then fatal( 'input buffer overflow');
		Dec( yycolno);
    Dec( yyposition);
		ungetbuf[ ungetbufptr] := c;
		Inc( ungetbufptr);
	end;


  Procedure put_char( c: Char);
  Begin
    With yyOutput Do
      If c = #0
      Then // ignore
      Else
        If c = nl
        Then Add('')
        Else Begin
          If Count = 0
          Then Add( '');
          yyoutput[ Count - 1] := yyoutput[ Count - 1] + c;
        End;
	End;

  Procedure yyStringOut( Const S: String);
  Var
    i : Integer;
  Begin
    For i := 1 To Length( S) Do
      put_char( S[ i]);
  End;

  Procedure yyStringOutFmt( Const aFormat: String; Const anArgs: Array Of Const);
  Begin
    yyStringOut( Format( aFormat, anArgs));
  End;


{
  Variables:

   Some state information is maintained to keep track with calls to yymore,
   yyless, reject, start and yymatch/yymark, and to initialize state
   information used by the lexical analyzer.

   yystext:

     contains the initial contents of the yytext variable; this will be the
     empty string, unless yymore is called which sets yystext to the current
     yytext.

   yysstate:

     start state of lexical analyzer (set to 0 during initialization, and
     modified in calls to the start routine).

   yylstate:

     line state information (1 if at beginning of line, 0 otherwise).

   yystack:

     stack containing matched rules; yymatches contains the number of matches.

   yypos:

     for each rule the last marked position (yymark); zeroed when rule has
     already been considered.

   yysleng:

     copy of the original yyleng used to restore state information when reject
     is used.
}

const

  max_matches = 1024;
  max_rules   = 256;

var

  yystext   : ShortString;
  yysstate  : Integer;
  yylstate  : Integer;
  yymatches : Integer;
  yystack   : array [ 1 .. max_matches] of Integer;
  yypos     : array [ 1 .. max_rules  ] of Integer;
  yysleng   : Integer;


// Utilities:

  Procedure echo;
  Var
    i : Integer;
  Begin
    For i := 1 To yyleng Do
      put_char( yytext[ i])
  End;


  Procedure yymore;
  Begin
    yystext := yytext;
  End;


  Procedure yyless( n: Integer);
  Var
    i : Integer;
  Begin
    For i := yyleng Downto n + 1 Do
      unget_char( yytext[ i]);
    SetLength( yyText, n);
    // yyleng := n;
  End;


  Procedure reject;
  Var
    i : Integer;
  Begin
    yyreject := true;
    For i := yyleng + 1 To yysleng Do
      yytext := yytext + get_char;
    dec( yymatches);
  end;


  Procedure return( n: Integer);
	Begin
    yyretval := n;
    yydone   := true;
  End;


  Procedure returnc( c: Char);
  Begin
    yyretval := ord( c);
		yydone   := true;
  End;


  Procedure start( state: Integer);
  Begin
    yysstate := state;
  End;


// yywrap:


  Function yywrap: Boolean;
  Begin
    If FileStack.Count > 0
    Then Begin
      FileStack.Pop;
      Result := yyinput_text = '';
    End
    Else Result := True;
	End;


  Procedure yyInclude( Const aFileName: String);
  Begin
    FileStack.Push( aFileName);
  End;

  Function  Printable( anInt: Integer): String;
  Const
    Ctrls : Array[ 0 .. 32] Of String[ 6] = (
      '^@ NUL', '^A SOH', '^B STX', '^C ETX', '^D EOT', '^E ENQ', '^F ACK', '^G BEL',
      '^H BS ', '^I HT ', '^J LF ', '^K VT ', '^L FF ', '^M CR ', '^N SO ', '^O SI ',
      '^P DLE', '^Q DC1', '^R DC2', '^S DC3', '^T DC4', '^U NAK', '^V SYN', '^W ETB',
      '^X CAN', '^Y EM ', '^Z SUB', '^[ ESC', '^\ FS ', '^] GS ', '^^ RS ', '^_ US ',
      '-- SP '
    );
  Var
    aByte : Byte;
  Begin
    If ( anInt >= 0) And ( anInt < 256)
    Then Begin
      aByte := anInt And $ff;
      Case aByte Of
        0 .. 32 : Result := Ctrls[ aByte];
        $7f     : Result := '-- DEL';
        Else      Result := Char( aByte);
      End;
    End
    Else Result := Format( '<%.4x>', [ anInt]);
  End;

  Procedure yyDebugOut( Const S: String);
  Var
    i : Integer;
  Begin
    If yyDebug
    Then Begin
      For i := 1 To Length( S) Do
        put_char( S[ i]);
      put_char( #10);
    End;
  End;

  Procedure yyDebugOutFmt( Const aFormat: String; Const anArgs: Array Of Const);
  Begin
    yyDebugOut( Format( aFormat, anArgs));
  End;


// Internal routines:


  Procedure yynew;
  Begin
    If yylastchar <> #0
    Then
      If yylastchar = nl
      Then yylstate := 1
      Else yylstate := 0;

    yystate   := yysstate + yylstate;
    yytext    := yystext;
    yystext   := '';
    yymatches := 0;
    yydone    := False;
	End;


  Procedure yyscan;
  Begin
    // If yyleng = 255
    // Then fatal( 'yytext overflow');
    yyactchar := get_char;
    SetLength( yyText, Length( yyText) + 1);
    // inc( yyleng);
		yytext[ yyleng] := yyactchar;
  End;


  Procedure yymark( n: Integer);
  Begin
    If n > max_rules
    Then fatal( 'too many rules');
    yypos[ n] := yyleng;
  End;


  Procedure yymatch( n: Integer);
  Begin
    inc( yymatches);
    If yymatches > max_matches
    Then fatal( 'match stack overflow');
    yystack[ yymatches] := n;
  End;


  Function yyfind( var n: Integer): Boolean;
  Begin
    yyreject := false;
    While ( yymatches > 0) And ( yypos[ yystack[ yymatches]] = 0) Do
      dec( yymatches);
    If yymatches>0
    Then Begin
      yysleng := yyleng;
      n       := yystack[ yymatches];
      yyless( yypos[ n]);
      yypos[ n] := 0;
      If yyleng > 0
      Then yylastchar := yytext[ yyleng]
      Else yylastchar := #0;
      yyfind := True;
    End
    Else Begin
      yyless( 0);
      yylastchar := #0;
      yyfind     := False;
    End
  End;


  Function yydefault: Boolean;
  Begin
		yyreject  := false;
		yyactchar := get_char;
		If yyactchar <> #0
    Then Begin
			put_char( yyactchar);
			yydefault := True;
    End
		Else Begin
      yylstate  := 1;
			yydefault := false;
    End;
		yylastchar := yyactchar;
	End;


  Procedure yyclear;
	Begin
		bufptr      := 0;
		ungetbufptr := 0;
		yysstate    := 0;
		yylstate    := 1;
		yylastchar  := #0;
		yytext      := '';
		yystext     := '';
	End;


  procedure yyinit;
  begin
    If Assigned( FileStack)
    Then FreeAndNil( FileStack);
    FileStack := TFileStack.Create;
    yyoutput.Clear;
    yyinput_text := '';
    yylineno     := 1;
    yycolno      := 0;
    yyposition   := 0;
    yyErrorCount := 0;
    yyclear;
  end;


initialization

  FileStack := Nil;
	yyoutput  := TStringList.Create;

finalization

	yyoutput.Free;
  If Assigned( FileStack)
  Then FreeAndNil( FileStack);

End.

